"use client"

import { useSearchParams } from "next/navigation"
import { useEffect, useState } from "react"
import ArticleCard from "@/components/article-card"
import { fetchArticles } from "@/lib/api"
import type { Article } from "@/lib/types"
import { Skeleton } from "@/components/ui/skeleton"

export default function ArticleList() {
  const searchParams = useSearchParams()
  const tagFilter = searchParams.get("tag")

  const [articles, setArticles] = useState<Article[]>([])
  const [loading, setLoading] = useState(true)
  const [error, setError] = useState<string | null>(null)

  useEffect(() => {
    const loadArticles = async () => {
      try {
        setLoading(true)
        const data = await fetchArticles(tagFilter)
        setArticles(data)
        setError(null)
      } catch (err) {
        console.error("Failed to fetch articles:", err)
        setError("Failed to load articles. Please try again later.")
      } finally {
        setLoading(false)
      }
    }

    loadArticles()
  }, [tagFilter])

  if (loading) {
    return <ArticleListSkeleton />
  }

  if (error) {
    return <div className="text-red-500">{error}</div>
  }

  if (articles.length === 0) {
    return (
      <div className="text-center py-12">
        <h3 className="text-xl font-medium text-muted-foreground">No articles found</h3>
        <p className="mt-2">Try changing your filter criteria</p>
      </div>
    )
  }

  return (
    <div className="grid grid-cols-1 md:grid-cols-2 gap-6">
      {articles.map((article) => (
        <ArticleCard key={article.id} article={article} />
      ))}
    </div>
  )
}

function ArticleListSkeleton() {
  return (
    <div className="grid grid-cols-1 md:grid-cols-2 gap-6">
      {Array.from({ length: 6 }).map((_, i) => (
        <div key={i} className="flex flex-col gap-2">
          <Skeleton className="h-48 w-full rounded-lg" />
          <Skeleton className="h-6 w-3/4" />
          <Skeleton className="h-4 w-full" />
          <Skeleton className="h-4 w-full" />
          <Skeleton className="h-4 w-2/3" />
        </div>
      ))}
    </div>
  )
}

