"use client"

import { useRouter, useSearchParams } from "next/navigation"
import { useEffect, useState } from "react"
import { fetchTags } from "@/lib/api"
import type { Tag } from "@/lib/types"
import { Button } from "@/components/ui/button"
import { Checkbox } from "@/components/ui/checkbox"
import { Label } from "@/components/ui/label"
import { Skeleton } from "@/components/ui/skeleton"

export function TagFilter() {
  const router = useRouter()
  const searchParams = useSearchParams()
  const currentTag = searchParams.get("tag")

  const [tags, setTags] = useState<Tag[]>([])
  const [loading, setLoading] = useState(true)
  const [error, setError] = useState<string | null>(null)

  useEffect(() => {
    const loadTags = async () => {
      try {
        setLoading(true)
        const data = await fetchTags()
        setTags(data)
        setError(null)
      } catch (err) {
        console.error("Failed to fetch tags:", err)
        setError("Failed to load tags. Please try again later.")
      } finally {
        setLoading(false)
      }
    }

    loadTags()
  }, [])

  const handleTagClick = (tagId: string) => {
    if (currentTag === tagId) {
      // If the tag is already selected, clear the filter
      router.push("/")
    } else {
      // Otherwise, filter by the selected tag
      router.push(`/articles?tag=${tagId}`)
    }
  }

  if (loading) {
    return (
      <div className="space-y-2">
        {Array.from({ length: 8 }).map((_, i) => (
          <Skeleton key={i} className="h-8 w-full" />
        ))}
      </div>
    )
  }

  if (error) {
    return <div className="text-red-500">{error}</div>
  }

  return (
    <div className="space-y-2">
      {tags.length === 0 ? (
        <p className="text-muted-foreground">No tags available</p>
      ) : (
        <>
          <div className="space-y-3">
            {tags.map((tag) => (
              <div key={tag.id} className="flex items-center space-x-2">
                <Checkbox
                  id={`tag-${tag.id}`}
                  checked={currentTag === tag.id}
                  onCheckedChange={() => handleTagClick(tag.id)}
                />
                <Label htmlFor={`tag-${tag.id}`} className="cursor-pointer">
                  {tag.name} {tag.count && <span className="text-muted-foreground">({tag.count})</span>}
                </Label>
              </div>
            ))}
          </div>

          {currentTag && (
            <Button variant="outline" size="sm" className="mt-4 w-full" onClick={() => router.push("/")}>
              Clear filters
            </Button>
          )}
        </>
      )}
    </div>
  )
}

